# pylint: disable=R0801
"""
Module that contains the class AbstractFrontend.
"""
import logging

from clinguin.utils import CustomArgs, Logger


class AbstractFrontend(CustomArgs):
    """
    Superclass from where every specialized frontend inherits from.
    This class defines the available elements.
    """

    def __init__(self, base_engine, args):
        self._args = args
        self._logger = logging.getLogger(Logger.client_logger_name)
        self._base_engine = base_engine

    @classmethod
    def available_syntax(cls, show_level):
        """
        Shows the available syntax for the chosen frontend.
        """
        print(show_level)
        return ""

    def window(self, cid, parent, attributes, callbacks):
        """
        Window placeholder.
        """
        print(
            "WINDOW: "
            + str(cid)
            + "::"
            + str(parent)
            + "::"
            + str(attributes)
            + "::"
            + str(callbacks)
        )

    def container(self, cid, parent, attributes, callbacks):
        """
        Container placeholder.
        """
        print(
            "CONTAINER: "
            + str(cid)
            + "::"
            + str(parent)
            + "::"
            + str(attributes)
            + "::"
            + str(callbacks)
        )

    def dropdown_menu(self, cid, parent, attributes, callbacks):
        """
        Dropdown-Menu placeholder.
        """
        print(
            "DROPDOWNMENU: "
            + str(cid)
            + "::"
            + str(parent)
            + "::"
            + str(attributes)
            + "::"
            + str(callbacks)
        )

    def dropdown_menu_item(self, cid, parent, attributes, callbacks):
        """
        Dropdown-menu-item placeholder.
        """
        print(
            "DROPDOWNMENUITEM: "
            + str(cid)
            + "::"
            + str(parent)
            + "::"
            + str(attributes)
            + "::"
            + str(callbacks)
        )

    def label(self, cid, parent, attributes, callbacks):
        """
        Label placeholder.
        """
        print(
            "LABEL: "
            + str(cid)
            + "::"
            + str(parent)
            + "::"
            + str(attributes)
            + "::"
            + str(callbacks)
        )

    def button(self, cid, parent, attributes, callbacks):
        """
        Button placeholder.
        """
        print(
            "BUTTON: "
            + str(cid)
            + "::"
            + str(parent)
            + "::"
            + str(attributes)
            + "::"
            + str(callbacks)
        )

    def menu_bar(self, cid, parent, attributes, callbacks):
        """
        Menu-Bar placeholder.
        """
        print(
            "MENUBAR: "
            + str(cid)
            + "::"
            + str(parent)
            + "::"
            + str(attributes)
            + "::"
            + str(callbacks)
        )

    def menu_bar_section(self, cid, parent, attributes, callbacks):
        """
        Menu-Bar-Section placeholder.
        """
        print(
            "MENUBARSECTION: "
            + str(cid)
            + "::"
            + str(parent)
            + "::"
            + str(attributes)
            + "::"
            + str(callbacks)
        )

    def menu_bar_section_item(self, cid, parent, attributes, callbacks):
        """
        Menu-Bar-Section-item placeholder.
        """
        print(
            "MENUBARSECTIONITEM: "
            + str(cid)
            + "::"
            + str(parent)
            + "::"
            + str(attributes)
            + "::"
            + str(callbacks)
        )

    def message(self, cid, parent, attributes, callbacks):
        """
        Message placeholder.
        """
        print(
            "MESSAGE: "
            + str(cid)
            + "::"
            + str(parent)
            + "::"
            + str(attributes)
            + "::"
            + str(callbacks)
        )

    def canvas(self, cid, parent, attributes, callbacks):
        """
        Canvas placeholder.
        """
        print(
            "CANVAS: "
            + str(cid)
            + "::"
            + str(parent)
            + "::"
            + str(attributes)
            + "::"
            + str(callbacks)
        )

    def draw_postprocessing(self, cid):
        """
        Draw postprocessing placeholder.
        """
        print("DRAW POSTPROCESSING" + "::" + str(cid))

    def draw(self, cid):
        """
        Draw placeholder.
        """
        print("DRAW" + "::" + str(cid))



"""
This module contains the EnumType.
"""

from enum import Enum

from .type import Type


class EnumType(Type, Enum):
    """
    The EnumType shall not be directly used, only subtypes of the EnumType
    (like the ChildLayout) shall be used.
    """



"""
Helper for the client-startup-process
"""

from clinguin.client import ClientBase
from clinguin.utils import Logger


def start(args):
    """
    Helper for the client-startup-process
    """
    Logger.setup_logger(args.log_args, process="client")

    client = ClientBase(args)
    client.start_up()



"""
Module that contains the AttributeNames class.
"""


class AttributeNames:  # pylint: disable=R0903
    """
    This class contains all the names for all the attributes that are used for the Frontend
    and also contains some standard descriptions of the attributes.
    """

    title = "title"
    message = "message"
    type = "type"

    label = "label"
    backgroundcolor = "background_color"
    foregroundcolor = "foreground_color"
    width = "width"
    height = "height"

    onhover = "on_hover"
    onhover_background_color = "on_hover_background_color"
    onhover_foreground_color = "on_hover_foreground_color"
    onhover_border_color = "on_hover_border_color"

    font_family = "font_family"
    font_size = "font_size"
    font_weight = "font_weight"

    resizable_x = "resizable_x"
    resizable_y = "resizable_y"

    child_layout = "child_layout"

    grid_column = "grid_column"
    grid_row = "grid_row"
    grid_column_span = "grid_column_span"
    grid_row_span = "grid_row_span"
    pos_x = "pos_x"
    pos_y = "pos_y"

    border_width = "border_width"
    border_color = "border_color"

    selected = "selected"
    fit_children_size = "fit"

    image = "image"
    image_path = "image_path"
    resize = "resize"

    flex_direction = "flex_direction"

    accelerator = "accelerator"
    descriptions = {
        title: "With this attribute one can set the title of a pop-up.",
        message: "With this attribute one can set the message of a pop-up.",
        type: "With this attribute one can set the look and feel of the pop-up.",
        label: "With this attribute one can set the label-text of the element.",
        backgroundcolor: "With this attribute one can define the background-color of the element.",
        foregroundcolor: "With this attribute one can set the foreground-color of the element.",
        width: "With this attribute one can set the width in pixels of the element.",
        height: "With this attribute one can set the height in pixels of the element.",
        onhover: "With this attribute one can enable or disable on-hover features for the element.",
        onhover_background_color: "With this attribute one can set the background color the element shall have, when"
        + "on_hover is enabled.",
        onhover_foreground_color: "With this attribute one can set the forground color the element shall have,"
        + "when on_hover is eneabled.",
        onhover_border_color: "With this attribute one can set the color the border of the element shall have,"
        + "when on_hover is enabled.",
        font_family: "With this attribute one can set the font-family of the element.",
        font_size: "With this attribute one can set the font size in pixels.",
        font_weight: "With this attribute one can set the font weight.",
        resizable_x: "With this attribute one can define if the element shall be resizable in x direction.",
        resizable_y: "With this attribute one can define if the element shall be resizable in y direction.",
        child_layout: "With this attribute one can define the layout of the children, i.e. how they are positioned.",
        grid_column: "With this attribute one can define in which column the element shall be positioned.",
        grid_row: "With this attribute one can define in which row the element shall be positioned.",
        grid_column_span: "With this attribute one can define, that the elements stretches over several columns.",
        grid_row_span: "With this attribute one can define, that the elements stretches over several rows.",
        pos_x: "With this attribute one sets the x-position of the element - it depends on the parent-child-layout"
        + "how this is defined (either pixels, relative as a percentage, ...).",
        pos_y: "With this attribute one sets the y-position of the element - it depends on the parent-child-layout"
        + "how this is defined (either pixels, relative as a percentage, ...).",
        border_width: "With this attribute one defines the width of the border in pixels.",
        border_color: "With this attribute one may set the border color.",
        selected: "With this attribute one may define the text that shall be displayed if something is selected.",
        fit_children_size: "[DEPRECATED] - With this attribute one can define that the element fits the children.",
        image: "Some backends define special values one can use, that e.g. a graph is shown (see corresponding"
        + "backend description).",
        image_path: "Include an image from a path.",
        resize: "With this attribute one can define if an image shall be resized.",
        accelerator: "The key binding for the menu option",
        flex_direction: "With this attribute one can set the ''direction'' (i.e., where it gets placed) of an"
        + "element which root has a specified flex layout.",
    }



"""
This module contains the StringType class.
"""

from .type import Type
from .utils.standard_text_processing import StandardTextProcessing


class StringType(Type):
    """
    The StringType shall be used, when the attribute value is string like.
    """

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = StandardTextProcessing.parse_string_with_quotes(parse_input)

        return parsed_string

    @classmethod
    def description(cls):
        return "String, can either be specified as a string or if it is simple as a symbol."



"""
This module contains various utils for tkinter-elements, which reduce code size, etc.
"""

from .attribute_names import AttributeNames
from .call_back_definition import CallBackDefinition
from .callback_names import CallbackNames
from .configure_border import ConfigureBorder
from .configure_font import ConfigureFont
from .configure_size import ConfigureSize
from .configure_text_element_size import ConfigureTextElementSize
from .extension_class import ExtensionClass
from .layout_controller import LayoutController
from .layout_follower import LayoutFollower

__all__ = [
    AttributeNames.__name__,
    CallbackNames.__name__,
    CallBackDefinition.__name__,
    LayoutFollower.__name__,
    ExtensionClass.__name__,
    LayoutController.__name__,
    ConfigureSize.__name__,
    ConfigureBorder.__name__,
    ConfigureFont.__name__,
    ConfigureTextElementSize.__name__,
]



"""
This package contains all tkinter_elements that are provided in clinguin.
"""

from .button import Button
from .canvas import Canvas
from .container import Container
from .dropdownmenu import Dropdownmenu
from .dropdownmenu_item import DropdownmenuItem
from .label import Label
from .menu_bar import MenuBar
from .menu_bar_section import MenuBarSection
from .menu_bar_section_item import MenuBarSectionItem
from .message import Message
from .root_cmp import RootCmp
from .window import Window

__all__ = [
    RootCmp.__name__,
    Window.__name__,
    Container.__name__,
    Dropdownmenu.__name__,
    DropdownmenuItem.__name__,
    Label.__name__,
    Button.__name__,
    MenuBar.__name__,
    MenuBarSection.__name__,
    MenuBarSectionItem.__name__,
    Message.__name__,
    Canvas.__name__,
]



"""
Module contains the menu bar section class.
"""

import tkinter as tk

from clinguin.utils.attribute_types import StringType

from ..tkinter_utils import AttributeNames
from .root_cmp import RootCmp


class MenuBarSection(RootCmp):
    """
    The menu bar section is a section of a menu bar
    (e.g. in the menu |main|contact|,
    where if one clicks on |contact|
    further the options |location|team|
    appear, a menu-bar-section would be |contact|,
    whereas |location| and |team| would be menu-bar-section-items.
    """

    def _init_element(self, elements):
        menubar_element = elements[self._parent].get_element()

        menubar_section = tk.Menu(menubar_element)

        return menubar_section

    @classmethod
    def _get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.label] = {
            "value": "default_label",
            "value_type": StringType,
        }

        return attributes

    def _set_sub_menu(self, elements):
        text = self._attributes[AttributeNames.label]["value"]

        menubar_element = elements[self._parent].get_element()
        menubar_element.add_cascade(label=text, menu=self._element)

    def _add_component_to_elements(self, elements):
        elements[str(self._id)] = self



"""
Annotations used to express overwrites and extensions of methods.
They enhance documentation and help to understand the code.
"""


def extends(super_cls):
    """
    Decorator to indicate that a method extends a method of a super class.
    Args:
        super_cls (_type_): The super class.
    """

    def decorator(func):
        def wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        wrapper.__name__ = func.__name__
        wrapper.__doc__ = (
            (func.__doc__ or "")
            + f"""
        Important:
            Extends :meth:`{super_cls.__name__}.{func.__name__}`
        """
        )
        return wrapper

    return decorator


def overwrites(super_cls):
    """
    Decorates a method to indicate that it overwrites a method of a super class.
    Args:
        super_cls (_type_): The super class.
    """

    def decorator(func):
        def wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        wrapper.__name__ = func.__name__
        wrapper.__doc__ = (
            (func.__doc__ or "")
            + f"""
        Important:
            Overwrites :meth:`{super_cls.__name__}.{func.__name__}`
        """
        )
        return wrapper

    return decorator



"""
Module that contains the ElementDto class.
"""

import json


class ElementDto:
    """
    The class that represents elements that are Json convertible,
    i.e. these components are the heart of the Json convertible hierarchy.
    """

    def __init__(self, cid, element_type, parent):
        self.id = str(cid)  # pylint: disable=C0103
        self.type = str(element_type)
        self.parent = str(parent)
        self.attributes = []
        self.when = []
        self.children = []

    def set_attributes(self, attributes):
        """
        Sets (all) the attributes of this element.
        """
        self.attributes = attributes

    def add_attribute(self, attribute):
        """
        Adds an attribute to this element.
        """
        self.attributes.append(attribute)

    def set_callbacks(self, callbacks):
        """
        Sets (all) the callbacks of this element.
        """
        self.when = callbacks

    def to_JSON(self):  # pylint: disable=C0103
        """
        Converts the element to a json.
        """
        return json.dumps(self, default=lambda o: str(o).__dict__)

    def clone(self):
        """
        Creates a new ElementDto object with the same properties as the self object.
        Note that all attributes, callbacks and children (elements) are deep copied/cloned.
        """
        clone = ElementDto(self.id, self.type, self.parent)

        cloned_attributes = []
        for attribute in self.attributes:
            cloned_attributes.append(attribute.clone())
        clone.set_attributes(cloned_attributes)

        cloned_callbacks = []
        for callback in self.when:
            cloned_callbacks.append(callback.clone())
        clone.set_callbacks(cloned_callbacks)

        cloned_children = []
        for child in self.children:
            cloned_children.append(child.clone())
        clone.children = cloned_children

        return clone

    def add_child(self, child):
        """
        Adds a child (element) to this element.
        """
        self.children.append(child)



"""
Module that contains the ClingoDL Backend.
"""

import textwrap

from clingo.ast import ProgramBuilder, parse_files
from clingo.script import enable_python

from clinguin.server.application.backends.clingo_backend import (
    ClingoBackend,
)
from clinguin.utils.annotations import extends

enable_python()
# pylint: disable=attribute-defined-outside-init


class TheoryBackend(ClingoBackend):
    """
    Abstract Backend that allows programs using a theory.
    It also includes the assignment in the domain state.
    """

    # ---------------------------------------------
    # Setups
    # ---------------------------------------------

    theory_class: type = type  # To be defined in subclasses, should be a callable class

    @extends(ClingoBackend)
    def _init_command_line(self):
        """
        Sets the dl configuration
        """
        super()._init_command_line()
        theory_config = (
            [a.split("=") for a in self._args.theory_config]
            if self._args.theory_config
            else []
        )
        self._theory_conf = [(a[0], a[1]) for a in theory_config]

    @extends(ClingoBackend)
    def _init_interactive(self):
        """
        Initializes the list of the assignments

        Attributes:
            _assignment (List[Tuple[str, int]]): The list of assignments
        """
        super()._init_interactive()
        self._assignment = []

    @extends(ClingoBackend)
    def _init_ds_constructors(self):
        super()._init_ds_constructors()
        self._add_domain_state_constructor("_ds_assign")

    @extends(ClingoBackend)
    def _create_ctl(self):
        """
        Registers the ClingconTheory.
        """
        super()._create_ctl()
        assert (
            self.theory_class is not None
        ), "theory_class must be defined in subclasses"
        self._theory = self.theory_class()
        for k, v in self._theory_conf:
            self._theory.configure(k, v)
        self._theory.register(self._ctl)

    @extends(ClingoBackend)
    def _load_file(self, f):
        """
        Uses the program builder to rewrite the theory atoms.
        """
        with ProgramBuilder(self._ctl) as bld:
            parse_files([f], lambda ast: self._theory.rewrite_ast(ast, bld.add))

    @extends(ClingoBackend)
    def _outdate(self):
        """
        Sets the assignment to empty.
        """
        super()._outdate()
        self._assignment = []

    # ---------------------------------------------
    # Solving
    # ---------------------------------------------
    @extends(ClingoBackend)
    def _prepare(self):
        """
        Prepares the theory before solving
        """
        # pylint: disable=attribute-defined-outside-init
        self._theory.prepare(self._ctl)

    @extends(ClingoBackend)
    def _on_model(self, model):
        """
        Sets the assignment from the model
        """
        super()._on_model(model)
        self._theory.on_model(model)
        # pylint: disable=attribute-defined-outside-init
        self._assignment = list(
            (key, val) for key, val in self._theory.assignment(model.thread_id)
        )

    # ---------------------------------------------
    # Class methods
    # ---------------------------------------------

    @classmethod
    @extends(ClingoBackend)
    def register_options(cls, parser):
        """
        Adds the `theory-config` option.
        """
        ClingoBackend.register_options(parser)

        parser.add_argument(
            "--theory-config",
            help=textwrap.dedent(
                """\
                    theory options list of <parameter>=<value>.
                 """
            ),
            nargs="*",
        )

    # ---------------------------------------------
    # Domain state
    # ---------------------------------------------

    @property
    def _ds_assign(self):
        """
        Adds program with assignments

        Includes predicate ``_clinguin_assign/2`` with the assignments.

        """
        if not self._ui_uses_predicate("_clinguin_assign", 2):
            return "% NOT USED\n"

        prg = ""
        for key, val in self._assignment:
            prg += f"_clinguin_assign({key},{val})."
        return prg



"""
Module that contains the BooleanType class.
"""

from .type import Type
from .utils.standard_text_processing import StandardTextProcessing


class BooleanType(Type):
    """
    Class that resembles a boolean value in the attributes.
    """

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = StandardTextProcessing.parse_string_with_quotes(parse_input)

        return_value = None

        if parsed_string.lower() == "true":
            return_value = True
        elif parsed_string.lower() == "false":
            return_value = False
        else:
            logger.error("Could not parse string to boolean: " + parsed_string)
            raise Exception("Could not parse string to boolean: " + parsed_string)

        return return_value

    @classmethod
    def description(cls):
        return (
            "For the boolean type, either true or false are allowed - either as string or as a clingo-symbol."
            + "If one provides it as a string, it is case-insensitive."
        )



import json

from reference_json_output import (
    BasicTest00,
    BasicTest01,
    BasicTest02,
    BasicTest03,
    BasicTest04,
    BasicTest05,
    Health,
)
from utils_test_utils import UtilsTestUtils


class TestBasic00_05:
    def setup_method(self, test_method):
        self.p, self.uvicorn_url = UtilsTestUtils.basic_tests_start_server(test_method)

    def teardown_method(self, test_method):
        UtilsTestUtils.shutdown_server(self.p)

    def test_health(self):
        uri = f"{self.uvicorn_url}/health"
        received_by_postman = json.dumps(Health.get_reference_json())
        print(received_by_postman)
        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_05(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest05.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_04(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest04.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_03_order_04(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest03.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

        data = '{"function":"add_assumption(p(1),true)"}'
        uri = f"{self.uvicorn_url}/backend"
        received_by_postman = json.dumps(BasicTest03.post_p_1_reference_json())

        UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

        data = '{"function":"clear_assumptions"}'
        uri = f"{self.uvicorn_url}/backend"
        received_by_postman = json.dumps(BasicTest03.post_p_3_reference_json())

        UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

    def test_basic_03_order_03(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest03.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

        data = '{"function":"clear_assumptions"}'
        uri = f"{self.uvicorn_url}/backend"
        received_by_postman = json.dumps(BasicTest03.post_p_2_reference_json())

        UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

    def test_basic_03_order_02(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest03.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

        data = '{"function":"add_assumption(p(1),true)"}'
        uri = f"{self.uvicorn_url}/backend"
        received_by_postman = json.dumps(BasicTest03.post_p_1_reference_json())

        UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

    def test_basic_03_order_01(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest03.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_02(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest02.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_01_order_02(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest01.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

        data = '{"function":"add_assumption(p(2),true)"}'
        uri = f"{self.uvicorn_url}/backend"
        received_by_postman = json.dumps(BasicTest01.post_p_2_reference_json())

        UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest01.get_p_2_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_01_order_01(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest01.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

        data = '{"function":"add_assumption(p(1),true)"}'
        uri = f"{self.uvicorn_url}/backend"
        received_by_postman = json.dumps(BasicTest01.post_p_1_reference_json())

        UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest01.get_p_1_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_00(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest00.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)



# pylint: disable=R0801
"""
Module contains the Container class.
"""
import tkinter as tk

from clinguin.utils.attribute_types import BooleanType, ColorType, SymbolType

from ..tkinter_utils import (
    AttributeNames,
    CallbackNames,
    ConfigureBorder,
    ConfigureSize,
    LayoutController,
    LayoutFollower,
)
from .root_cmp import RootCmp


class Container(
    RootCmp, LayoutFollower, LayoutController, ConfigureSize, ConfigureBorder
):
    """
    The container is a generic element which can be used for layouting,
    hovering effects or even callbacks. Generally it is recommended to use
    it as a ''container'' for multiple other elements, e.g. labels, buttons, etc.
    """

    def _init_element(self, elements):
        container = tk.Frame(elements[str(self._parent)].get_element())
        return container

    @classmethod
    def _get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.backgroundcolor] = {
            "value": "white",
            "value_type": ColorType,
            "description": "CUSTOM-BACKGROUND-COLOR-DESCRIPTION <- Now normal:"
            + AttributeNames.descriptions[AttributeNames.backgroundcolor],
        }
        # Interactive-Attributes
        attributes[AttributeNames.onhover] = {"value": False, "value_type": BooleanType}
        attributes[AttributeNames.onhover_background_color] = {
            "value": "",
            "value_type": ColorType,
        }
        attributes[AttributeNames.onhover_border_color] = {
            "value": "",
            "value_type": ColorType,
        }

        return attributes

    @classmethod
    def _get_callbacks(cls, callbacks=None):
        if callbacks is None:
            callbacks = {}

        callbacks["click"] = {"operation": None, "operation_type": SymbolType}

        return callbacks

    # ----------------------------------------------------------------------------------------------
    # -----Attributes----
    # ----------------------------------------------------------------------------------------------

    def _set_background_color(self, elements, key=AttributeNames.backgroundcolor):
        self._logger.debug(str(elements))
        value = self._attributes[key]["value"]
        self._element.configure(background=value)

    def _set_on_hover(self, elements):
        on_hover = self._attributes[AttributeNames.onhover]["value"]
        on_hover_color = self._attributes[AttributeNames.onhover_background_color][
            "value"
        ]
        on_hover_border_color = self._attributes[AttributeNames.onhover_border_color][
            "value"
        ]

        if on_hover:

            def enter(event):
                self._logger.debug(str(event))
                if on_hover_color != "":
                    self._set_background_color(
                        elements, key=AttributeNames.onhover_background_color
                    )
                if on_hover_border_color != "":
                    self._set_border_background_color(
                        elements, key=AttributeNames.onhover_border_color
                    )

            def leave(event):
                self._logger.debug(str(event))
                self._set_background_color(elements, key=AttributeNames.backgroundcolor)
                self._set_border_background_color(
                    elements, key=AttributeNames.border_color
                )

            self._element.bind("<Enter>", enter)
            self._element.bind("<Leave>", leave)

    # ----------------------------------------------------------------------------------------------
    # -----Actions----
    # ----------------------------------------------------------------------------------------------

    def _define_click_event(self, elements, key=CallbackNames.click):
        self._logger.debug(str(elements))
        if self._callbacks[key] and self._callbacks[key]["operation"]:

            def dropdown_menu_item_click(event):
                self._logger.debug(str(event))
                self._base_engine.assume(self._callbacks[key]["operation"])

            self._element.bind("<Button-1>", dropdown_menu_item_click)



# pylint: disable=R0903
"""
Module that contains the ElementDao.
"""
from clorm import Predicate, RawField


class ElementDao(Predicate):
    """
    Elements define building blocks of the UI.

    - id: Identifies the element for further references.
    - type: The type of element (window, container, button etc).
    - parent: The id of the parent element. The identifier root is used as the root element of the UI.

    """

    # pylint: disable=abstract-method

    id = RawField
    type = RawField
    parent = RawField

    class Meta:
        """
        Meta class
        """

        name = "elem"



# pylint: disable=R0801
# pylint: disable=too-many-lines
"""
Module that contains the ClingoBackend.
"""
import functools
import logging
import textwrap
import time
import tempfile
import base64
from functools import cached_property
from pathlib import Path
from typing import Any
from importlib.resources import files

from clingo import Control, parse_term
from clingo.script import enable_python

from clinguin.server import StandardJsonEncoder, UIState
from clinguin.server.data.domain_state import solve, tag

from ....utils.logger import domctl_log
from ....utils.transformer import UsesSignatureTransformer

enable_python()


# pylint: disable=attribute-defined-outside-init
# pylint: disable=too-many-public-methods
class ClingoBackend:
    """
    This backend contains the basic clingo functionality for a backend using clingo.
    """

    def __init__(self, args):
        """
        Creates the Backend with the given arguments.
        It will setup all attributes by calling :func:`~_init_ds_constructors()` and :func:`~_restart()`.

        Generally this method should NOT be overwritten by custom backends.
        Instead, custom backends should overwrite specialized methods.

        Arguments:
                args (ArgumentParser): The arguments from the argument parser that are given for the registered options.

        """
        self._args = args

        self._logger = logging.getLogger(args.log_args["name"])

        # Setup static attributes that might be changed by custom backends and must be preserved after restarts
        self._init_ds_constructors()

        # Restart the backend to initialize all attributes
        self._restart()

    # ---------------------------------------------
    # Class methods
    # ---------------------------------------------

    @classmethod
    def register_options(cls, parser):
        """
        Registers options in the command line.

        It can be extended by custom backends to add custom command-line options.

        Arguments:
                parser (ArgumentParser): A group of the argparse argument parser

        Example:

                .. code-block:: python

                        @classmethod
                        def register_options(cls, parser):
                                ClingoBackend.register_options(parser)

                                parser.add_argument(
                                        "--my-custom-option",
                                        help="Help message",
                                        nargs="*",
                                )

        """
        parser.add_argument(
            "--domain-files",
            nargs="+",
            help="Files with the domain specific encodings and the instances",
            metavar="",
        )
        parser.add_argument(
            "--ui-files",
            nargs="+",
            help="Files with the encodings that generate the UI predicates: elem, attr and when",
            metavar="",
        )
        parser.add_argument(
            "-c",
            "--const",
            action="append",
            help="Constant passed to clingo, <id>=<term> replaces term occurrences of <id> with <term>",
            metavar="",
        )
        parser.add_argument(
            "--clingo-ctl-arg",
            action="append",
            help="""Argument that will be passed to clingo control object for the domain.
                Should have format <name>=<value>, for example parallel-mode=2 will become --parallel-mode=2.""",
            metavar="",
        )
        parser.add_argument(
            "--default-opt-mode",
            type=str,
            help="Default optimization mode for computing a model",
            default="ignore",
            metavar="",
        )

        parser.add_argument(
            "--opt-timeout",
            help=textwrap.dedent(
                """\
                    Optional timeout for searching for optimal models.
                    The timeout is not exactly enforced (might take longer)
                    but only checked after each solution is found.
                 """
            ),
            type=int,
            metavar="",
        )

        parser.add_argument(
            "--optional-files",
            nargs="+",
            help="Optional domain files that can be activated/deactivated",
            metavar="",
        )
        parser.add_argument(
            "--explicit-show",
            help="""If set, the UI encoding will only consider atoms that are shown in the domain encoding.
            Otherwise, all atoms are considered. While this option can improve performance,
            the user must make sure any predicate used in the UI is shown in the domain encoding.""",
            action="store_true",
        )

        parser.add_argument(
            "--include-ds-in-response",
            help="""If set, the response of the server will include the domain state as a dictionary.
            Setting it will impact performance.""",
            action="store_true",
        )

    # ---------------------------------------------
    # Properties
    # ---------------------------------------------

    @property
    def _is_browsing(self):
        """
        Property to tell if clinguin is in browsing mode.
        """
        return self._iterator is not None

    @property
    def _constants_argument_list(self) -> list:
        """
        Gets the constants as a list of strings in the format "-c <name>=<value>"
        """
        return [f"-c {k}={v}" for k, v in self._constants.items()]

    @property
    def _ctl_arguments_list(self) -> list:
        """
        Gets the list of arguments used for creating a control object
        """
        return (
            ["0"]
            + self._constants_argument_list
            + [f"--{o}" for o in self._clingo_ctl_arg]
        )

    @property
    def _assumption_list(self) -> list:
        """
        A list of assumptions in the format of form (a, True) or (a, False)
        """
        return self._assumptions

    # ---------------------------------------------
    # Initialization
    # ---------------------------------------------

    def _restart(self):
        """
        Restarts the backend by setting all attributes,
        initializing controls and grounding.
        It is automatically called when the server starts.

        See Also:
                :func:`~_init_command_line`, :func:`~_init_interactive`,
                :func:`~_outdate`, :func:`~_init_ctl`, :func:`~_ground`
        """
        self._init_command_line()
        self._init_interactive()
        self._outdate()
        self._init_ctl()
        self._ground()

    def _init_ds_constructors(self):
        """
        This method initializes the domain state constructors list and the backup cache dictionary.
        It also adds the default domain state constructors to the list.
        This method is called only when the server starts.

        Attributes:
                _domain_state_constructors (list): A list to store the domain state constructors.
                _backup_ds_cache (dict): A dictionary to store the backup domain state cache.


        It can be extended by custom backends to add/edit domain state constructors.
        Adding a domain state constructor should be done by calling :func:`~_add_domain_state_constructor()`.

        Example:

                .. code-block:: python

                        @property
                        def _ds_my_custom_constructor(self):
                                # Creates custom program
                                return "my_custom_program."

                        def _init_ds_constructors(self):
                                super()._init_ds_constructors()
                                self._add_domain_state_constructor("_ds_my_custom_constructor")

        """
        self._domain_state_constructors = []
        self._backup_ds_cache = {}
        self._add_domain_state_constructor("_ds_context")
        self._add_domain_state_constructor("_ds_constants")
        self._add_domain_state_constructor("_ds_browsing")
        self._add_domain_state_constructor("_ds_cautious_optimal")
        self._add_domain_state_constructor("_ds_brave_optimal")
        self._add_domain_state_constructor("_ds_cautious")
        self._add_domain_state_constructor("_ds_brave")
        self._add_domain_state_constructor("_ds_model")  # Keep after brave and cautious
        self._add_domain_state_constructor("_ds_opt")
        self._add_domain_state_constructor("_ds_unsat")  # Keep after all solve calls
        self._add_domain_state_constructor("_ds_atom")
        self._add_domain_state_constructor("_ds_assume")
        self._add_domain_state_constructor("_ds_external")
        self._add_domain_state_constructor("_ds_files")

    def _init_command_line(self):
        """
        Initializes the attributes based on the command-line arguments provided.
        This method is called when the server starts or after a restart.

        Attributes:

                _domain_files (list): The list of domain files provided via command line.
                _ui_files (list): The list of UI files provided via command line.
                _constants (dict): The dictionary of constants provided via command line.
                _clingo_ctl_arg (list): The list of clingo control arguments provided via command line.

        If any command line arguments are added in :meth:`~ClingoBackend.register_options`,
        they should be initialized here.

        Example:

                .. code-block:: python

                                def _init_command_line(self):
                                        super()._init_command_line()
                                        self._my_custom_attr = self._args.my_custom_option


        """

        self._domain_files = self._args.domain_files or []
        self._optional_files = self._args.optional_files or []
        self._original_domain_files = self._domain_files.copy()

        if not self._args.ui_files:
            self._logger.warning(
                "UI files need to be provided under --ui-files. Using default UI encoding."
            )
            default_ui_path = files(
                "clinguin.server.application.backends.standard_utils"
            ).joinpath("default_ui.lp")
            self._ui_files = [default_ui_path]
        else:
            self._ui_files = self._args.ui_files

        self._constants = {}
        if self._args.const is not None:
            for c in self._args.const:
                c = c.strip(" ")
                if "=" not in c:
                    raise ValueError("Invalid constant format. Expected name=value.")
                name, value = c.split("=")
                self._constants[name] = value

        self._clingo_ctl_arg = self._args.clingo_ctl_arg or []

        self._default_opt_mode = self._args.default_opt_mode

        self._opt_timeout = self._args.opt_timeout

        self._explicit_show = self._args.explicit_show

        self._include_ds_in_response = self._args.include_ds_in_response

    def _init_interactive(self):
        """
        Initializes the attributes that will change during the interaction.
        This method is called when the server starts or after a restart.

        Attributes:
                _context (list): A list to store the context set by the general handler of requests.
                _handler (clingo.SolveHandle): The handler set while browsing in the `next_solution` operation.
                _iterator (iter): The iterator set while browsing in the `next_solution` operation.
                _ctl (clingo.Control): The domain control set in `_init_ctl`.
                _ui_state (:class:`UIState`): A UIState object used to handle the UI construction,
                set in every call to `_update_ui_state`.
                _atoms (set[str]): A set to store the atoms set dynamically in operations during the interaction.
                _assumptions (set[(str,bool)]): A set to store the assumptions set dynamically in operations during the
                interaction.
                _externals (dict): A dictionary with true, false and released sets of external atoms
                _model (list[clingo.Symbol]): The model set in `on_model`.
                _unsat_core (list[int]): The unsatisfiable core set in `on_model`.
                _cost (list): A list to store the cost set in `on_model`.
                _optimal (bool): A boolean indicating if the solution is optimal, set in `on_model`.
                _optimizing (bool): A boolean indicating if the solver is currently optimizing, set in `on_model`.
                _messages (list[tuple[str,str,str]]): A list to store the messages (title, content, type) to be shown in
                the UI, set dynamically in operations during the interaction.
        """
        # Context: Set by the general handler of requests
        self._context = []

        # Domain Control: Set in _init_ctl
        self._ctl = None

        # UIState object to handle the UI construction: Set in every time in _update_ui_state
        self._ui_state = None

        # Atoms and assumptions: Set dynamically in operations during the interaction
        self._atoms = set()
        self._assumptions = set()
        self._externals = {"true": set(), "false": set(), "released": set()}

        # Handler and Iterator: Set while browsing in next_solution operation
        self._handler = None
        self._iterator = None

        # Attributes from the model: Set in on_model
        self._model = None
        self._unsat_core = None
        self._cost = []
        self._optimal = False
        self._optimizing = False

        # Messages to be shown in the UI: Set dynamically in operations during the interaction
        self._messages = []

        self._uploaded_files = {}
        self._active_files = []

    def _init_ctl(self):
        """
        Creates the domain control and loads the domain files.

        See Also:
                :func:`~_create_ctl`, :func:`~_load_and_add`
        """
        self._create_ctl()
        self._load_and_add()

    def _create_ctl(self) -> None:
        """
        Initializes the control object (domain-control).
        It is used when the server is started or after a restart.

        See Also:
                :func:`~_load_file`
        """
        self._logger.debug(
            domctl_log(f"domainctl = Control({self._ctl_arguments_list})")
        )
        self._ctl = Control(self._ctl_arguments_list)

    def _load_and_add(self) -> None:
        """
        Loads domain files and atoms into the control.

        This method iterates over the domain files and atoms specified in the instance and loads them into the control.
        It raises an exception if a domain file does not exist or if there is a syntax error in the logic program file.

        Raises:
                Exception: If a domain file does not exist or if there is a syntax error in the logic program file.

        See Also:
                :func:`~_load_file`
        """
        for f in self._domain_files:
            path = Path(f)
            if not path.is_file():
                self._logger.critical("File %s does not exist", f)
                raise Exception(f"File {f} does not exist")

            try:
                self._load_file(f)
            except Exception as e:
                self._logger.critical(
                    "Failed to load file %s (there is likely a syntax error in this logic program file).",
                    f,
                )
                self._logger.critical(str(e))
                raise e

        for atom in self._atoms:
            self._logger.debug(domctl_log('domctl.add("base", [], {str(atom)} + ".")'))
            self._ctl.add("base", [], str(atom) + ".")

    def _load_file(self, f):
        """
        Loads a file into the control. This method can be overwritten if any pre-processing is needed.

        Arguments:
                f (str): The file path
        """
        self._logger.debug(domctl_log(f"domctl.load({str(f)})"))
        self._ctl.load(str(f))

    def _outdate(self):
        """
        Outdates all the dynamic values when a change has been made.
        Any current interaction in the models wil be terminated by canceling the search and removing the iterator.

        See Also:
                :func:`~_clear_cache`
        """
        if self._handler:
            self._handler.cancel()
            self._handler = None
        self._iterator = None
        self._model = None
        self._clear_cache()

    # ---------------------------------------------
    # Setters
    # ---------------------------------------------

    def _add_domain_state_constructor(self, method: str):
        """
        Adds a method name to the domain constructors.
        The provided method needs to be annotated with ``@property`` or ``@cached_property``

        Arguments:
                method (str): Name of the property method
        """

        self._domain_state_constructors.append(method)

    def _set_context(self, context):
        """
        Sets the context. Used by general endpoint handler after a request.

        Arguments:
                context: The context dictionary
        """
        self._context = context

    def _set_constant(self, name: str, value: Any) -> None:
        """
        Sets a constant in the backend and restarts the control.

        Args:
                name (str): name of the constant
                value (Any): value of the constant
        """
        name = name.strip('"')
        value = str(value).strip('"')
        self._constants[name] = value

        self._logger.debug("Constant %s updated successfully to %s", name, value)

    def _add_atom(self, predicate_symbol):
        """
        Adds an atom if it hasn't been already added

        Arguments:
                predicate_symbol (clingo.Symbool): The symbol for the atom
        """
        if predicate_symbol not in self._atoms:
            self._atoms.add(predicate_symbol)

    def _set_external(self, symbol, name):
        """
        Sets the external value of a symbol.

        Args:
                symbol (clingo.Symbol): The clingo symbol to be set
                name (str): Either "true", "false" or "release"
        """
        name = name.strip('"')
        if name == "release":
            self._logger.debug(domctl_log(f"ctl.release_external({symbol})"))
            self._ctl.release_external(symbol)
            self._externals["released"].add(symbol)

            if symbol in self._externals["true"]:
                self._externals["true"].remove(symbol)

            if symbol in self._externals["false"]:
                self._externals["false"].remove(symbol)

        elif name == "true":
            self._logger.debug(domctl_log(f"ctl.assign_external({symbol}, True)"))
            self._ctl.assign_external(symbol, True)
            self._externals["true"].add(symbol)

            if symbol in self._externals["false"]:
                self._externals["false"].remove(symbol)

            if symbol in self._externals["released"]:
                self._externals["released"].remove(symbol)

        elif name == "false":
            self._logger.debug(domctl_log(f"ctl.assign_external({symbol}, False)"))
            self._ctl.assign_external(symbol, False)
            self._externals["false"].add(symbol)

            if symbol in self._externals["true"]:
                self._externals["true"].remove(symbol)

            if symbol in self._externals["released"]:
                self._externals["released"].remove(symbol)

        else:
            raise ValueError(
                f"Invalid external value {name}. Must be true, false or relase"
            )

    def _add_assumption(self, symbol, value="true"):
        """
        Adds an assumption to the list of assumptions.

        Args:
                symbol (clingo.Symbol): The clingo symbol to be added as a True assumption
                value (true): The value of the assumption either "true" or "false"
        """
        bool_val = value == "true"
        self._assumptions.add((symbol, bool_val))

    # ---------------------------------------------
    # Solving
    # ---------------------------------------------

    def _ground(self, program="base", arguments=None):
        """
        Grounds the provided program

        Arguments:
                program (str): The name of the program to ground (defaults to "base")
                arguments (list, optional): The list of arguments to ground the program. Defaults to an empty list.
        """
        arguments = arguments or []
        arguments = [arguments] if not isinstance(arguments, list) else arguments
        self._logger.debug(domctl_log(f"domctl.ground([({program}, {arguments})])"))
        arguments_symbols = [parse_term(a) for a in arguments]
        self._ctl.ground([(program, arguments_symbols)])
        self._load_externals()

    def _load_externals(self):
        """
        Loads the externals into the control.
        This method is called after grounding to ensure that the externals are loaded.
        """
        for symbol in self._ctl.symbolic_atoms:
            if symbol.is_external:
                symbol_str = str(symbol.symbol)
                self._logger.debug("Setting external symbol found: %s", symbol_str)
                self._externals["released"].add(symbol_str)

    def _prepare(self):
        """
        Does any preparation before a solve call.
        """

    def _on_model(self, model):
        """
        This method is called each time a model is obtained by the domain control.
        It sets the model, and optimization attributes.
        It can be extended to add custom features of the model.

        Arguments:
                model (clingo.Model): The found clingo model
        """
        self._optimizing = len(model.cost) > 0
        if len(model.cost) > 0:
            self._logger.debug("       Cost: %s", model.cost)
        self._optimal = model.optimality_proven
        self._cost = model.cost

    # ---------------------------------------------
    # UI state
    # ---------------------------------------------

    def _update_ui_state(self):
        """
        Updates the UI state by calling all domain state methods
        and creating a new control object (ui_control) using the UI files provided
        """
        domain_state = self._domain_state
        self._ui_state = UIState(
            self._ui_files, domain_state, self._constants_argument_list
        )
        self._ui_state.update_ui_state()
        self._ui_state.replace_images_with_b64()
        for m in self._messages:
            self._ui_state.add_message(m[0], m[1], m[2])
        self._messages = []

    # ---------------------------------------------
    # Domain state
    # ---------------------------------------------

    def _clear_cache(self, methods=None):
        """
        Clears the cache of domain state constructor methods

        Arguments:
                methods (list, optional): A list with the methods to remove the cache from.
                        If no value is passed then all cache is removed
        """
        if methods is None:
            methods = self._domain_state_constructors
        for m in methods:
            if m in self.__dict__:
                self._backup_ds_cache[m] = self.__dict__[m]
                del self.__dict__[m]

    def _call_solver_with_cache(
        self, ds_id: str, ds_tag: str, models: int, opt_mode: str, enum_mode: str
    ):
        """
        Generic function to call the using exiting cache on browsing.
        Un UNSAT it returns the output saved in the cache

        Arguments:
                ds_id: Identifier used in the cache
        Returns:
                The program tagged
        """
        if self._is_browsing:
            self._logger.debug("Returning cache for %s", ds_id)
            return (
                self._backup_ds_cache[ds_id] if ds_id in self._backup_ds_cache else ""
            )
        self._logger.debug(domctl_log(f'domctl.configuration.solve.models = {models}"'))
        self._logger.debug(
            domctl_log(f'domctl.configuration.solve.opt_mode = {opt_mode}"')
        )
        self._logger.debug(
            domctl_log(f'domctl.configuration.solve.enum_mode = {enum_mode}"')
        )
        self._ctl.configuration.solve.models = models
        self._ctl.configuration.solve.opt_mode = opt_mode
        self._ctl.configuration.solve.enum_mode = enum_mode
        self._prepare()
        self._logger.debug(
            domctl_log(
                f"domctl.solve({[(str(a), b) for a, b in self._assumption_list]}, yield_=True)"
            )
        )
        symbols, ucore = solve(
            self._ctl,
            self._assumption_list,
            self._on_model,
            include_all_atoms=not self._explicit_show,
        )
        self._unsat_core = ucore
        if symbols is None:
            self._logger.warning("Got an UNSAT result with the given domain encoding.")
            return (
                self._backup_ds_cache[ds_id] if ds_id in self._backup_ds_cache else ""
            )
        return " ".join([str(s) + "." for s in list(tag(symbols, ds_tag))]) + "\n"

    @functools.lru_cache(maxsize=None)  # pylint: disable=[method-cache-max-size-none]
    def _ui_uses_predicate(self, name: str, arity: int):
        """
        Returns a truth value of weather the ui_files contain the given signature.

        Args:
                name (str): Predicate name
                arity (int): Predicate arity
        """
        transformer = UsesSignatureTransformer(name, arity)
        self._logger.debug("Transformer parsing UI files to find %s/%s", name, arity)
        transformer.parse_files(self._ui_files)
        if not transformer.contained:
            self._logger.debug(
                "Predicate NOT contained. Domain constructor will be skipped"
            )
        return transformer.contained

    @property
    def _domain_state(self):
        """
        Gets the domain state by calling all the domain constructor methods

        Some domain state constructors might skip the computation if the UI does not require them.
        """
        ds = ""
        for f in self._domain_state_constructors:
            ds += f"\n%%%%%%%% {f} %%%%%%%\n"
            ds += getattr(self, f)
        return ds

    @property
    def _domain_state_dict(self):
        """
        Gets the domain state as a dictionary for the response of the server.

        Some domain state constructors might skip the computation if the UI does not require them.
        """
        ds = {}
        for f in self._domain_state_constructors:
            prg = getattr(self, f)
            ctl = Control(["0"])
            ctl.add("base", [], prg)
            ctl.ground([("base", [])])
            symbols = []
            with ctl.solve(yield_=True) as handle:
                for m in handle:
                    symbols = [str(a).replace('"', "'") for a in m.symbols(shown=True)]
            ds[f] = symbols
        return ds

    @property
    def _ds_files(self):
        """Domain state for file management"""
        prg = "#defined _clinguin_optional_file/1. #defined _clinguin_active_file/1. "

        # Optional files
        for filepath in self._optional_files:
            name = Path(filepath).name
            prg += f'_clinguin_optional_file("{name}"). '

        for name in self._uploaded_files:
            prg += f'_clinguin_optional_file("{name}"). '

        # Active files
        for filepath in self._active_files:
            if filepath in self._uploaded_files.values():
                name = next(k for k, v in self._uploaded_files.items() if v == filepath)
            else:
                name = Path(filepath).name
            prg += f'_clinguin_active_file("{name}"). '

        return prg + "\n"

    # -------- Domain state methods

    @property
    def _ds_context(self):
        """
        Adds context information from the client.

        Includes predicate  ``_clinguin_context/2`` indicating each key and value in the context.
        """
        prg = "#defined _clinguin_context/2. "
        for a in self._context:
            value = str(a.value)
            try:
                symbol = parse_term(value)
            except Exception:
                symbol = None
            if symbol is None:
                value = f'"{value}"'
            prg += f"_clinguin_context({str(a.key)},{value})."
        return prg + "\n"

    @cached_property
    def _ds_model(self):
        """
        Computes model and adds all atoms as facts.
        When the model is being iterated by the user, the current model is returned.
        It will use as optimality the mode set in the command line as `default-opt-mode` (`ignore` by default).

        It uses a cache that is erased after an operation makes changes in the control.
        """
        if self._model is None:
            self._logger.debug(
                domctl_log('domctl.configuration.solve.enum_mode = "auto"')
            )
            self._ctl.configuration.solve.models = 1
            self._ctl.configuration.solve.opt_mode = self._default_opt_mode
            self._ctl.configuration.solve.enum_mode = "auto"

            self._prepare()
            self._logger.debug(
                domctl_log(
                    f"domctl.solve({[(str(a), b) for a, b in self._assumption_list]}, yield_=True)"
                )
            )

            symbols, ucore = solve(
                self._ctl,
                self._assumption_list,
                self._on_model,
                include_all_atoms=not self._explicit_show,
            )
            self._unsat_core = ucore
            if symbols is None:
                self._logger.warning(
                    "Got an UNSAT result with the given domain encoding."
                )
                return (
                    self._backup_ds_cache["_ds_model"]
                    + "\n".join([str(a) + "." for a in self._atoms])
                    if "_ds_model" in self._backup_ds_cache
                    else ""
                )
            self._model = symbols

        tagged = (
            " ".join(["_clinguin_model(" + str(s) + ")." for s in self._model]) + "\n"
        )
        return " ".join([str(s) + "." for s in self._model]) + "\n" + tagged

    @cached_property
    def _ds_brave(self):
        """
        Computes brave consequences adds them as predicates ``_any/1``.
        This are atoms that appear in some model.
        If it is not used in the UI files then the computation is not performed.

        It uses a cache that is erased after an operation makes changes in the control.
        """
        if not self._ui_uses_predicate("_any", 1):
            return "% NOT USED\n"

        return self._call_solver_with_cache("_ds_brave", "_any", 0, "ignore", "brave")

    @cached_property
    def _ds_cautious(self):
        """
        Computes cautious consequences adds them as predicates ``_all/1``.
        This are atoms that appear in all models.
        If it is not used in the UI files then the computation is not performed.

        It uses a cache that is erased after an operation makes changes in the control.
        """
        if not self._ui_uses_predicate("_all", 1):
            return "% NOT USED\n"

        return self._call_solver_with_cache(
            "_ds_cautious", "_all", 0, "ignore", "cautious"
        )

    @cached_property
    def _ds_brave_optimal(self):
        """
        Computes brave consequences for only optimal solutions adds them as predicates ``_any_opt/1``.
        This are atoms that appear in some optimal model.
        If it is not used in the UI files then the computation is not performed.

        It uses a cache that is erased after an operation makes changes in the control.
        """
        if not self._ui_uses_predicate("_any_opt", 1):
            return "% NOT USED\n"

        return self._call_solver_with_cache(
            "_ds_brave_optimal", "_any_opt", 0, "optN", "brave"
        )

    @cached_property
    def _ds_cautious_optimal(self):
        """
        Computes cautious consequences of optimal models adds them as predicates ``_all_opt/1``.
        This are atoms that appear in all optimal models.
        If it is not used in the UI files then the computation is not performed.

        It uses a cache that is erased after an operation makes changes in the control.
        """
        if not self._ui_uses_predicate("_all_opt", 1):
            return "% NOT USED\n"

        return self._call_solver_with_cache(
            "_ds_cautious_optimal", "_all_opt", 0, "optN", "cautious"
        )

    @property
    def _ds_unsat(self):
        """
        Adds information about the statisfiablity of the domain control

        Includes predicate ``_clinguin_unsat/0`` if the domain control is unsat
        """
        prg = "#defined _clinguin_unsat/0. "
        if self._unsat_core is not None:
            prg += "_clinguin_unsat."
        return prg + "\n"

    @property
    def _ds_browsing(self):
        """
        Adds information about the browsing state

        Includes predicate  ``_clinguin_browsing/0`` if the user is browsing solutions
        """
        prg = "#defined _clinguin_browsing/0. "
        if self._is_browsing:
            prg += "_clinguin_browsing."
        return prg + "\n"

    @property
    def _ds_atom(self):
        """
        Adds information about the added atoms.

        Includes predicate  ``_clinguin_atom/1`` for every atom that was added.
        """
        prg = "#defined _clinguin_atom/1. "
        for a in self._atoms:
            prg += f"_clinguin_atom({str(a)}). "
        return prg + "\n"

    @property
    def _ds_assume(self):
        """
        Adds information about the assumptions.

        Includes predicate  ``_clinguin_assume/2`` for every atom that was assumed,
        where the second argument is either true or false.
        """
        prg = "#defined _clinguin_assume/2. "
        for a, v in self._assumption_list:
            v_str = "true" if v else "false"
            prg += f"_clinguin_assume({str(a)},{v_str}). "
        return prg + "\n"

    @property
    def _ds_external(self):
        """
        Adds information about the external atoms

        Includes predicate  ``_clinguin_external/2`` for every external atom that has been set.
        """
        prg = "#defined _clinguin_external/2. "
        for a in self._externals["true"]:
            prg += f"_clinguin_external({str(a)},true). "
        for a in self._externals["false"]:
            prg += f"_clinguin_external({str(a)},false). "
        for a in self._externals["released"]:
            prg += f"_clinguin_external({str(a)},release). "
        return prg + "\n"

    @property
    def _ds_opt(self):
        """
        Adds program to pass with optimality information.

        Includes predicates:
         - ``_clinguin_cost/1``: With a single tuple indicating the cost of the current model
         - ``_clinguin_cost/2``: With the index and cost value, linearizing predicate ``_clinguin_cost/1``
         - ``_clinguin_optimal/0``: If the solution is optimal
         - ``_clinguin_optimizing/0``: If there is an optimization in the program
        """
        prg = "#defined _clinguin_cost/2. #defined _clinguin_cost/1."
        prg += "#defined _clinguin_optimal/0. #defined _clinguin_optimizing/0."

        for i, c in enumerate(self._cost):
            prg += f"_clinguin_cost({i},{c}). "
        if self._optimal:
            prg += "_clinguin_optimal. "
        if self._optimizing:
            prg += "_clinguin_optimizing. "

        prg += f"_clinguin_cost({tuple(self._cost)}).\n"
        return prg

    @property
    def _ds_constants(self):
        """
        Adds constants  values.

        Includes predicate ``_clinguin_const/2`` for each constant provided
        in the command line and used in the domain files.
        """
        prg = "#defined _clinguin_const/2. "
        for k, v in self._constants.items():
            prg += f'_clinguin_const("{k}",{v}).'
        return prg + "\n"

    ########################################################################################################

    # ---------------------------------------------
    # File management helpers
    # ---------------------------------------------

    def _find_file_path(self, name):
        """Find file path by name"""
        name = name.strip('"')
        if name in self._uploaded_files:
            return self._uploaded_files[name]
        for f in self._optional_files:
            if Path(f).name == name:
                return f
        return None

    def _reload_domain(self):
        """Reload domain after file changes"""
        self._domain_files = self._original_domain_files + self._active_files
        self._outdate()
        self._init_ctl()
        self._ground()

    # ---------------------------------------------
    # Public operations
    # ---------------------------------------------

    def get(self):
        """
        Updates the UI and transforms the facts into a JSON.
        This method will be automatically called after executing all the operations.
        """
        self._update_ui_state()
        ds = None
        if self._include_ds_in_response:
            self._logger.debug("Including domain state in response")
            ds = self._domain_state
        json_structure = StandardJsonEncoder.encode(self._ui_state, ds)
        return json_structure

    def restart(self):
        """
        Restarts the backend. It will initialize all attributes, remove atoms, assumptions and externals,
        restart the control object by initializing all parameters, controls, ending the browsing and grounding.
        """
        self._restart()

    def ground(self, program, arguments=None):
        """
        Grounds the given program. Notice that the base program is grounded when the server starts.
        This operation can be used for grounding encodings with multiple programs in a multi-shot setting.

        Arguments:
                program (str): The name of the program to ground used in the #program directive
                arguments (tuple, optional): The list of arguments to ground the program. Defaults to an empty list.
                These are the arguments of your #program directive.
                For instance, in #program step(t). the argument is t.
        """
        self._ground(program, arguments)

    def update(self):
        """
        Updates the UI by clearing the cache and computing the models again.
        """
        self._clear_cache()

    def download(self, show_prg=None, file_name="clinguin_download.lp"):
        """
        Downloads the current model. It must be selected first via :func:`~select` .

        Arguments:
                show_prg (_type_, optional): Program to filter output using show statements. Defaults to None.
                file_name (str, optional): The name of the file for the download. Defaults to "clinguin_download.lp".
        """
        if self._model is None:
            raise RuntimeError("Cant download when there is no model")
        show_prg = show_prg or ""
        prg = "\n".join([f"{s}." for s in self._model])
        ctl = Control()
        ctl.add("base", [], prg)
        try:
            ctl.add("base", [], show_prg.replace('"', ""))
        except RuntimeError as exc:
            raise Exception(
                "Show program can't be parsed. Make sure it is a valid clingo program."
            ) from exc
        ctl.ground([("base", [])])
        with ctl.solve(yield_=True) as hnd:
            for m in hnd:
                atoms = [f"{str(s)}." for s in m.symbols(shown=True)]

        final_prg = "\n".join(atoms)

        file_name = file_name.strip('"')
        with open(file_name, "w", encoding="UTF-8") as file:
            file.write(final_prg)
        self._messages.append(
            (
                "Download successful",
                f"Information saved in file {file_name}.",
                "success",
            )
        )

    def clear_atoms(self):
        """
        Removes all atoms and resets the backend.
        and finally updates the model and returns the updated gui as a Json structure.
        """
        self._outdate()
        self._atoms = set()
        self._init_ctl()
        self._ground()

    def add_atom(self, predicate):
        """
        Adds an atom, restarts the control and grounds

        Arguments:

                predicate (str): The clingo symbol to be added
        """
        predicate_symbol = parse_term(predicate)
        if predicate_symbol not in self._atoms:
            self._add_atom(predicate_symbol)
            self._init_ctl()
            self._ground()
            self._outdate()

    def remove_atom(self, predicate):
        """
        Removes an atom (if present), restarts the control and grounds again

        Arguments:

                predicate (str): The clingo symbol to be added
        """
        predicate_symbol = parse_term(predicate)
        if predicate_symbol in self._atoms:
            self._atoms.remove(predicate_symbol)
            self._init_ctl()
            self._ground()
            self._outdate()

    def next_solution(self, opt_mode="ignore", opt_timeout=None):
        """
        Obtains the next solution. If a no browsing has been started yet, then it calls solve,
        otherwise it iterates the models in the last call. To keep the atoms shown in the solution, use :func:`~select`.

        Arguments:
                opt_mode: The clingo optimization mode, bu default is 'ignore', to browse only optimal models use 'optN'
                opt_timeout: The timeout in seconds to find an optimal solution. If not provided, the default timeout
                        provided in the command line is used. If no timeout is provided, it will wait until an optimal
                        solution is found. Otherwise, if the timeout is reached, it will return the current solution
                        without proving optimality.
        """
        opt_timeout = int(opt_timeout) if opt_timeout is not None else None
        if self._ctl.configuration.solve.opt_mode != opt_mode:
            self._logger.debug("Ended browsing since opt mode changed")
            self._outdate()
        optimizing = opt_mode in ["optN", "opt"]
        if not self._iterator:
            self._logger.debug(
                domctl_log(f"domctl.configuration.solve.opt_mode = {opt_mode}")
            )
            self._ctl.configuration.solve.enum_mode = "auto"
            self._ctl.configuration.solve.opt_mode = opt_mode
            self._ctl.configuration.solve.models = 0

            self._prepare()
            self._logger.debug(
                domctl_log(
                    f"domctl.solve({[(str(a), b) for a, b in self._assumption_list]}, yield_=True)"
                )
            )
            self._handler = self._ctl.solve(self._assumption_list, yield_=True)

            self._iterator = iter(self._handler)
        try:
            start = time.time()
            model = next(self._iterator)
            self._clear_cache(["_ds_model"])
            self._on_model(model)
            self._model = model.symbols(
                shown=True, atoms=not self._explicit_show, theory=True
            )
            while optimizing and not model.optimality_proven:
                if len(model.cost) == 0:
                    self._messages.append(
                        ("Browsing Warning", "No optimization provided", "warning", 1)
                    )
                    self._logger.warning(
                        "No optimization statement provided in encoding but optimization condition provided"
                        " in 'next_solution' operation. Exiting browsing."
                    )
                    break
                timeout = opt_timeout if opt_timeout is not None else self._opt_timeout
                if timeout is not None and time.time() - start > timeout:
                    self._logger.warning(
                        "Timeout for finding optimal model was reached. Returning model without proving optimality."
                    )
                    break
                self._logger.debug("Skipping non-optimal model!")
                model = next(self._iterator)
                self._clear_cache(["_ds_model"])
                self._on_model(model)

            self._model = model.symbols(
                shown=True, atoms=not self._explicit_show, theory=True
            )
        except StopIteration:
            if optimizing:
                m = "No more optimal solutions"
            else:
                m = "No more solutions"

            self._logger.info(m)
            self._outdate()
            self._messages.append(("Browsing Information", m, "info"))

    def clear_assumptions(self):
        """
        Removes all assumptions.
        """
        # pylint: disable=attribute-defined-outside-init
        self._outdate()
        self._assumptions = set()

    def add_assumption(self, atom, value="true"):
        """
        Adds an atom `a` as an assumption.

        If the value is "true", the atom is assumed to be true.
        This assumption can be considered as an integrity constraint:
        ``:- not a.`` forcing the program to entail the given atom.

        If the value is "false", the atom is assumed to be false:
        This assumption can be considered as an integrity constraint:
        ``:- a.`` forcing the program to never entail the given atom.

        Notice that the assumption must be generated by your domain encoding.

        Arguments:

                atom (str): The clingo symbol to be added as a true assumption
                value (str): The value of the assumption either "true" or "false"
        """
        atom_symbol = parse_term(atom)
        if atom_symbol not in [a[0] for a in self._assumptions]:
            self._add_assumption(atom_symbol, value)
            self._outdate()
        else:
            self._logger.warning("Assumption already exists. Not adding it again.")

    def remove_assumption(self, atom):
        """
        Removes an atom from the assumptions list regardless of its value.
        This will allow the atom to take any value in the solution.

        Arguments:
                atom (str): The clingo symbol to be removed
        """
        atom_symbol = parse_term(atom)
        for a, v in self._assumptions:
            if a == atom_symbol:
                self._assumptions.remove((a, v))
                self._outdate()
                return

    def remove_assumption_signature(self, atom):
        """
        Removes from the list of assumptions those matching the given atom.
        Unlike function remove_assumption, this one allows for partial matches using the
        placeholder constant `any`.
        This will allow the atom to take any value in the solution.

        Arguments:

                atom (str): The atom description as a symbol,
                        where the reserver word `any` is used to state that anything can
                        take part of that position. For instance, `person(anna,any)`,
                        will remove all assumptions of atom person, where the first argument is anna.
        """
        atom_symbol = parse_term(atom)
        arity = len(atom_symbol.arguments)
        to_remove = []
        for s, v in self._assumptions:
            if s.match(atom_symbol.name, arity):
                for i, a in enumerate(atom_symbol.arguments):
                    if str(a) != "any" and s.arguments[i] != a:
                        break
                else:
                    to_remove.append((s, v))
                    continue
        for a in to_remove:
            self._assumptions.remove(a)
        if len(to_remove) > 0:
            self._outdate()

    def set_external(self, atom, value):
        """
        Sets the value of an external. Externals must be defined in the domain files using `#external`.
        The truth value of external atoms can then be provided by the user via this function.

        Arguments:

                atom (str): The clingo symbol to be set
                value (str): The value (release, true or false)
        """
        symbol = parse_term(atom)
        name = value
        self._outdate()

        self._set_external(symbol, name)

    def select(self, show_prg: str = ""):
        """
        Select the current solution during browsing.
        All atoms in the solution are added as assumptions in the backend.

        Arguments:

                show_program (str): An optional show program to filter atoms

        """
        if self._model is None:
            self._messages.append(
                ("No solution", "There is no solution to be selected", "danger")
            )
            self._logger.error(
                "No solution. No model has been computed that can be selected"
            )
        else:
            symbols_to_ignore = self._externals["true"]
            symbols_to_ignore.union(self._externals["false"])
            if show_prg == "":
                model = self._model
            else:
                model = []
                ctl = Control(["--warn=none"])
                try:
                    ctl.add("base", [], show_prg.strip('"'))
                except RuntimeError as exc:
                    raise Exception(
                        "Show program can't be parsed. Make sure it is a valid clingo program."
                    ) from exc
                prg = "\n".join([f"{str(s)}." for s in self._model])
                ctl.add("base", [], prg)
                ctl.ground([("base", [])])

                def add_shown(m):
                    for s in m.symbols(shown=True):
                        model.append(s)

                ctl.solve(on_model=add_shown)
            for s in model:  # pylint: disable=E1133
                if s not in symbols_to_ignore:
                    self._add_assumption(s, "true")
        self._outdate()

    def stop_browsing(self):
        """
        Stops the current browsing.
        """
        self._outdate()

    def set_constant(self, name: str, value: Any):
        """
        Sets a constant value. Will reinitialize the control, ground and set arguments
        """
        self._set_constant(name, value)
        self._init_interactive()
        self._outdate()
        self._init_ctl()
        self._ground()

    def upload_file(self):
        """
        Upload file using the context. The context should have the name of the file under `filename` and the
        file content in base64 under `filecontent`.
        """
        filename = next((c.value for c in self._context if c.key == "filename"), None)
        content = next((c.value for c in self._context if c.key == "filecontent"), None)
        if not filename:
            raise RuntimeError("No file name found in context")

        if not content:
            raise RuntimeError("No content found in context")

        original_name = next(
            (c.value for c in self._context if c.key == "_filename"), filename
        )
        original_name = original_name.strip('"')

        existing_files = {Path(f).name for f in self._optional_files} | set(
            self._uploaded_files.keys()
        )
        unique_name = original_name
        counter = 1
        while unique_name in existing_files:
            base, ext = Path(original_name).stem, Path(original_name).suffix
            unique_name = f"{base}_{counter}{ext}"
            counter += 1

        decoded_content = base64.b64decode(content).decode("utf-8")
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".lp") as f:
            f.write(decoded_content)
            self._uploaded_files[unique_name] = f.name

        self._messages.append(("", f"Uploaded file {unique_name}", "success"))

    def activate_file(self, filename):
        """
        Activate an optional file to include it in the domain control.

        Args:
                filename (str): The name of the file to activate.
        """
        path = self._find_file_path(filename)
        if not path:
            self._logger.error(
                "File not found: %s, available files: %s",
                filename,
                self._optional_files,
            )
            self._messages.append(("Error", f"File not found: {filename}", "danger"))
            return
        if path not in self._active_files:
            self._active_files.append(path)
            self._reload_domain()

    def deactivate_file(self, filename):
        """
        Deactivate an active file, removing it from the active files list.

        Args:
                filename (str): The name of the file to deactivate.
        """
        path = self._find_file_path(filename)
        if path in self._active_files:
            self._active_files.remove(path)
            self._reload_domain()

    def remove_file(self, filename):
        """
                Remove a file completely.

        Args:
                filename (str): The name of the file to remove.
        """
        name = filename.strip('"')
        path = self._find_file_path(name)
        self._active_files = [p for p in self._active_files if p != path]
        self._optional_files = [f for f in self._optional_files if Path(f).name != name]
        if name in self._uploaded_files:
            Path(self._uploaded_files.pop(name)).unlink(missing_ok=True)
        self._reload_domain()
        self._messages.append(("", f"Removed file {name}", "success"))



"""
Module that contains the backends
"""

# pylint: disable=cyclic-import
from clinguin.server.application.backends.clingo_backend import ClingoBackend
from clinguin.server.application.backends.clingodl_backend import ClingoDLBackend
from clinguin.server.application.backends.clingcon_backend import ClingconBackend
from clinguin.server.application.backends.clingraph_backend import ClingraphBackend
from clinguin.server.application.backends.fclingo_backend import FclingoBackend
from clinguin.server.application.backends.explanation_backend import ExplanationBackend

__all__ = [
    ClingoBackend.__name__,
    ClingraphBackend.__name__,
    ExplanationBackend.__name__,
    ClingoBackend.__name__,
    ClingoDLBackend.__name__,
    ClingconBackend.__name__,
    FclingoBackend.__name__,
]



# pylint: disable=R0801,R1716
"""
Contains the window class.
"""
import tkinter as tk

from clinguin.utils.attribute_types import ChildLayoutType, ColorType, IntegerType

from ..tkinter_utils import AttributeNames, LayoutController
from .root_cmp import RootCmp


class Window(RootCmp, LayoutController):
    """
    The window class is the top-lvl. component, therefore it MUST be used in every clinguin application!
    For available attributes see syntax definition.
    """

    def _init_element(self, elements):
        window = tk.Tk()
        window.title("Clinguin")

        return window

    @classmethod
    def _get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.backgroundcolor] = {
            "value": "white",
            "value_type": ColorType,
        }
        attributes[AttributeNames.width] = {"value": 0, "value_type": IntegerType}
        attributes[AttributeNames.height] = {"value": 0, "value_type": IntegerType}
        attributes[AttributeNames.pos_x] = {"value": -1, "value_type": IntegerType}
        attributes[AttributeNames.pos_y] = {"value": -1, "value_type": IntegerType}
        attributes[AttributeNames.resizable_x] = {"value": 1, "value_type": IntegerType}
        attributes[AttributeNames.resizable_y] = {"value": 1, "value_type": IntegerType}

        return attributes

    def _set_background_color(
        self, elements, key=AttributeNames.backgroundcolor
    ):  # pylint: disable=W0613
        value = self._attributes[key]["value"]
        self._element.configure(background=value)

    def _set_dimensions(self, elements):  # pylint: disable=W0613
        width = self._attributes[AttributeNames.width]["value"]
        height = self._attributes[AttributeNames.height]["value"]

        pos_x = self._attributes[AttributeNames.pos_x]["value"]
        pos_y = self._attributes[AttributeNames.pos_y]["value"]

        if height > 0 and width > 0:
            child_layout_value = self._attributes[AttributeNames.child_layout]["value"]

            if child_layout_value in (
                ChildLayoutType.FLEX,
                ChildLayoutType.RELSTATIC,
                ChildLayoutType.ABSSTATIC,
            ):
                self._element.pack_propagate(0)
            elif child_layout_value == ChildLayoutType.GRID:
                self._element.grid_propagate(0)

            if pos_x < 0:
                pos_x = int((int(self._element.winfo_screenwidth()) - int(width)) / 2)
            if pos_y < 0:
                pos_y = int((int(self._element.winfo_screenheight()) - int(height)) / 2)

            self._element.geometry(
                str(width) + "x" + str(height) + "+" + str(pos_x) + "+" + str(pos_y)
            )

        elif (height > 0 and width <= 0) or (height <= 0 and width > 0):
            self._logger.warning(
                "For the tkinter window one must set both height and width to positive values (not just one)."
            )

    def _set_resizable(self, elements):  # pylint: disable=W0613
        self._element.resizable(
            self._attributes[AttributeNames.resizable_x]["value"],
            self._attributes[AttributeNames.resizable_y]["value"],
        )

    def _add_component_to_elements(self, elements):
        elements[str(self._id)] = self



"""
Contains the Message class.
"""

import tkinter as tk
from tkinter import messagebox

# This import is used implicitly
from clinguin.utils.attribute_types import PopupTypesType, StringType

from ..tkinter_utils import AttributeNames
from .root_cmp import RootCmp


class Message(RootCmp):
    """
    A message is a pop up, which has a type a title and a message.
    """

    def _init_element(self, elements):
        message = tk.Message(elements[str(self._parent)].get_element())
        return message

    @classmethod
    def _get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.type] = {
            "value": PopupTypesType.INFO,
            "value_type": PopupTypesType,
        }
        attributes[AttributeNames.title] = {"value": "", "value_type": StringType}
        attributes[AttributeNames.message] = {"value": "", "value_type": StringType}

        return attributes

    def _set_values(self, elements):
        self._logger.debug(str(elements))
        attr_type = self._attributes[AttributeNames.type]["value"]
        title = self._attributes[AttributeNames.title]["value"]
        message = self._attributes[AttributeNames.message]["value"]
        if PopupTypesType.INFO == attr_type:
            messagebox.showinfo(title=title, message=message)
        elif PopupTypesType.WARNING == attr_type:
            messagebox.showwarning(title=title, message=message)
        elif attr_type in ["danger", PopupTypesType.ERROR]:
            messagebox.showerror(title=title, message=message)
        else:
            self._logger.warning("Cannot display popup-type %s", attr_type)



"""
Clinguin package - package entry point
"""

import copy
import sys
import threading
from datetime import datetime

from .client_helper import start as client_start
from .parse_input import ArgumentParser
from .server_helper import start as server_start


def args_to_dict_converter(args_dict, timestamp, name_prefix=""):
    """
    Converts the ''external-logger-config'' representation into an internal one
    """
    log_dict = {}
    log_dict["file_enabled"] = args_dict[name_prefix + "log_enable_file"]
    log_dict["shell_disabled"] = args_dict[name_prefix + "log_disable_shell"]
    log_dict["name"] = args_dict[name_prefix + "logger_name"]
    log_dict["level"] = args_dict[name_prefix + "log_level"]
    log_dict["format_shell"] = args_dict[name_prefix + "log_format_shell"]
    log_dict["format_file"] = args_dict[name_prefix + "log_format_file"]
    log_dict["timestamp"] = timestamp

    return log_dict


def main():
    """
    Main entry point into the project
    """

    parser = ArgumentParser()

    if len(sys.argv) > 1:
        process = sys.argv[1]
    else:
        process = sys.argv[0]

    args = parser.parse(process, sys.argv[1:])

    args_dict = vars(args)

    timestamp = datetime.now().strftime("%Y-%m-%d::%H:%M:%S")

    if args.process == "server":
        log_dict = args_to_dict_converter(args_dict, timestamp)

        args_copy = copy.deepcopy(args)
        args_copy.log_args = log_dict
        server = threading.Thread(target=server_start, args=[args_copy])

        server.start()

    elif args.process == "client":
        log_dict = args_to_dict_converter(args_dict, timestamp)

        args_copy = copy.deepcopy(args)
        args_copy.log_args = log_dict

        client_start(args_copy)

    elif args.process == "client-server":
        server_log_dict = args_to_dict_converter(
            args_dict, timestamp, name_prefix="server_"
        )
        args_copy = copy.deepcopy(args)
        args_copy.log_args = server_log_dict

        server = threading.Thread(target=server_start, args=[args_copy])

        server.start()
        server.join(2)

        client_log_dict = args_to_dict_converter(
            args_dict, timestamp, name_prefix="client_"
        )

        args_copy = copy.deepcopy(args)
        args_copy.log_args = client_log_dict

        client_start(args_copy)



"""
Module that contains the EndpointsHelper class.
"""

import logging

from ...utils import CaseConverter, Logger


class EndpointsHelper:
    """
    The EndpointsHelper class is responsible for getting the correct method in the ''backend''
    (here backend refers to ClingoBackend, TemporalBackend, etc.). This is done via reflections,
    i.e. it checks if the method is in the backend, if yes the method it returned.
    """

    @classmethod
    def call_function(cls, backend, name, args, kwargs):
        """
        Helper function that calls given a backend, a name for a function/method and arguments,
        the respective function/method.
        """
        logger = logging.getLogger(Logger.server_logger_name)

        found = False
        function = None

        snake_case_name = name
        camel_case_name = CaseConverter.snake_case_to_camel_case(snake_case_name)

        public_functions = [
            str(func) for func in dir(backend) if not func.startswith("_")
        ]
        if snake_case_name.startswith("_"):
            error_string = "Cannot call private functions '" + name + "' in backend."
            logger.error(error_string)
            logger.error("Available operations: \n\t%s", "\n\t".join(public_functions))
            raise Exception(error_string)

        if hasattr(backend, snake_case_name):
            function = getattr(backend, snake_case_name)
            found = True
        elif hasattr(backend, camel_case_name):
            function = getattr(backend, camel_case_name)
            found = True

        if found:
            result = function(*args, **kwargs)
            return result
        error_string = "Could not find operation '" + name + "' in backend."
        logger.error(error_string)
        logger.error("Available operations: \n\t%s", "\n\t".join(public_functions))
        raise Exception(error_string)
